"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const globby = require("globby");
const os_1 = __importDefault(require("os"));
const vscode_uri_1 = require("vscode-uri");
const forest_1 = require("./forest");
const imports_1 = require("./imports");
const astProvider_1 = require("./providers/astProvider");
const codeActionProvider_1 = require("./providers/codeActionProvider");
const codeLensProvider_1 = require("./providers/codeLensProvider");
const completionProvider_1 = require("./providers/completionProvider");
const definitionProvider_1 = require("./providers/definitionProvider");
const diagnosticsProvider_1 = require("./providers/diagnostics/diagnosticsProvider");
const elmAnalyseDiagnostics_1 = require("./providers/diagnostics/elmAnalyseDiagnostics");
const elmMakeDiagnostics_1 = require("./providers/diagnostics/elmMakeDiagnostics");
const documentFormatingProvider_1 = require("./providers/documentFormatingProvider");
const documentSymbolProvider_1 = require("./providers/documentSymbolProvider");
const foldingProvider_1 = require("./providers/foldingProvider");
const hoverProvider_1 = require("./providers/hoverProvider");
const referencesProvider_1 = require("./providers/referencesProvider");
const renameProvider_1 = require("./providers/renameProvider");
const workspaceSymbolProvider_1 = require("./providers/workspaceSymbolProvider");
const documentEvents_1 = require("./util/documentEvents");
const utils = __importStar(require("./util/elmUtils"));
const textDocumentEvents_1 = require("./util/textDocumentEvents");
class ElmWorkspace {
    constructor(elmWorkspace, connection, settings, parser) {
        this.elmWorkspace = elmWorkspace;
        this.connection = connection;
        this.settings = settings;
        this.parser = parser;
        this.forest = new forest_1.Forest();
        this.connection.console.info(`Starting language server for folder: ${this.elmWorkspace}`);
        this.imports = new imports_1.Imports(parser);
        this.documentEvents = new documentEvents_1.DocumentEvents(this.connection, this.elmWorkspace);
        this.textDocumentEvents = new textDocumentEvents_1.TextDocumentEvents(this.documentEvents);
    }
    registerInitializedProviders() {
        return __awaiter(this, void 0, void 0, function* () {
            // tslint:disable:no-unused-expression
            new astProvider_1.ASTProvider(this.connection, this.forest, this.documentEvents, this.imports, this.parser);
            new foldingProvider_1.FoldingRangeProvider(this.connection, this.forest);
            new completionProvider_1.CompletionProvider(this.connection, this.forest, this.imports);
            new hoverProvider_1.HoverProvider(this.connection, this.forest, this.imports);
            new definitionProvider_1.DefinitionProvider(this.connection, this.forest, this.imports);
            new referencesProvider_1.ReferencesProvider(this.connection, this.forest, this.imports);
            new documentSymbolProvider_1.DocumentSymbolProvider(this.connection, this.forest);
            new workspaceSymbolProvider_1.WorkspaceSymbolProvider(this.connection, this.forest);
            new codeLensProvider_1.CodeLensProvider(this.connection, this.forest, this.imports);
            new renameProvider_1.RenameProvider(this.connection, this.forest, this.imports);
        });
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            const settings = yield this.settings.getClientSettings();
            const documentFormatingProvider = new documentFormatingProvider_1.DocumentFormattingProvider(this.connection, this.elmWorkspace, this.textDocumentEvents, this.settings);
            const elmAnalyse = settings.elmAnalyseTrigger !== "never"
                ? new elmAnalyseDiagnostics_1.ElmAnalyseDiagnostics(this.connection, this.elmWorkspace, this.textDocumentEvents, this.settings, documentFormatingProvider)
                : null;
            const elmMake = new elmMakeDiagnostics_1.ElmMakeDiagnostics(this.connection, this.elmWorkspace, this.settings);
            // tslint:disable:no-unused-expression
            new diagnosticsProvider_1.DiagnosticsProvider(this.connection, this.elmWorkspace, this.settings, this.textDocumentEvents, elmAnalyse, elmMake);
            new codeActionProvider_1.CodeActionProvider(this.connection, elmAnalyse, elmMake);
            yield this.initWorkspace();
        });
    }
    initWorkspace() {
        return __awaiter(this, void 0, void 0, function* () {
            let elmVersion;
            try {
                elmVersion = yield utils.getElmVersion(yield this.settings.getClientSettings(), this.elmWorkspace, this.connection);
            }
            catch (e) {
                this.connection.console.warn(`Could not figure out elm version, this will impact how good the server works. \n ${e.stack}`);
            }
            try {
                const path = `${this.elmWorkspace.fsPath}elm.json`;
                this.connection.console.info(`Reading elm.json from ${path}`);
                // Find elm files and feed them to tree sitter
                const elmJson = require(path);
                const type = elmJson.type;
                const elmFolders = new Map();
                if (type === "application") {
                    elmJson["source-directories"].forEach((folder) => __awaiter(this, void 0, void 0, function* () {
                        elmFolders.set(this.elmWorkspace.fsPath + folder, true);
                    }));
                }
                else {
                    elmFolders.set(`${this.elmWorkspace.fsPath}src`, true);
                }
                elmFolders.set(`${this.elmWorkspace.fsPath}tests`, true);
                this.connection.console.info(`${elmFolders.size} source-dirs and test folders found`);
                const elmHome = this.findElmHome();
                const packagesRoot = `${elmHome}/${elmVersion}/${this.packageOrPackagesFolder(elmVersion)}/`;
                const dependencies = type === "application"
                    ? Object.assign({}, elmJson.dependencies.direct, elmJson["test-dependencies"].direct) : Object.assign({}, elmJson.dependencies, elmJson["test-dependencies"]);
                if (type === "application") {
                    for (const key in dependencies) {
                        if (dependencies.hasOwnProperty(key)) {
                            const maintainer = key.substring(0, key.indexOf("/"));
                            const packageName = key.substring(key.indexOf("/") + 1, key.length);
                            const pathToPackageWithVersion = `${packagesRoot}${maintainer}/${packageName}/${dependencies[key]}/src`;
                            elmFolders.set(pathToPackageWithVersion, false);
                        }
                    }
                }
                else {
                    for (const key in dependencies) {
                        if (dependencies.hasOwnProperty(key)) {
                            const maintainer = key.substring(0, key.indexOf("/"));
                            const packageName = key.substring(key.indexOf("/") + 1, key.length);
                            const pathToPackage = `${packagesRoot}${maintainer}/${packageName}/`;
                            const allVersionFolders = fs_1.readdirSync(pathToPackage, "utf8").map(folderName => {
                                return {
                                    version: folderName,
                                    versionPath: `${pathToPackage}${folderName}`,
                                };
                            });
                            // TODO Actually honor the version constraints here
                            const matchedFolder = utils.findDepVersion(allVersionFolders, dependencies[key]);
                            const pathToPackageWithVersion = matchedFolder
                                ? `${matchedFolder.versionPath}/src`
                                : `${allVersionFolders[allVersionFolders.length - 1].versionPath}/src`;
                            elmFolders.set(pathToPackageWithVersion, false);
                        }
                    }
                }
                const elmFilePaths = this.findElmFilesInFolders(elmFolders);
                this.connection.console.info(`Found ${elmFilePaths.length.toString()} files to add to the project`);
                for (const filePath of elmFilePaths) {
                    this.connection.console.info(`Adding ${filePath.path.toString()}`);
                    const fileContent = fs_1.readFileSync(filePath.path.toString(), "utf8");
                    let tree;
                    tree = this.parser.parse(fileContent);
                    this.forest.setTree(vscode_uri_1.URI.file(filePath.path).toString(), filePath.writable, true, tree);
                }
                this.forest.treeIndex.forEach(item => {
                    this.connection.console.info(`Adding imports ${item.uri.toString()}`);
                    this.imports.updateImports(item.uri, item.tree, this.forest);
                });
                this.connection.console.info("Done parsing all files.");
            }
            catch (error) {
                this.connection.console.error(error.stack);
            }
        });
    }
    findElmFilesInFolders(elmFolders) {
        let elmFilePaths = [];
        for (const element of elmFolders) {
            elmFilePaths = elmFilePaths.concat(this.findElmFilesInFolder(element));
        }
        return elmFilePaths;
    }
    findElmFilesInFolder(element) {
        return globby
            .sync(`${element[0]}/**/*.elm`)
            .map(path => ({ path, writable: element[1] }));
    }
    packageOrPackagesFolder(elmVersion) {
        return elmVersion === "0.19.0" ? "package" : "packages";
    }
    findElmHome() {
        const elmHomeVar = process.env.ELM_HOME;
        if (elmHomeVar) {
            return elmHomeVar;
        }
        return utils.isWindows
            ? `${os_1.default.homedir()}/AppData/Roaming/elm`
            : `${os_1.default.homedir()}/.elm`;
    }
}
exports.ElmWorkspace = ElmWorkspace;
//# sourceMappingURL=elmWorkspace.js.map